"""Main class."""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck
import functools
import itertools
import logging
import os
import re
from collections.abc import Callable, Sequence
from copy import copy
from typing import Any, Protocol

from filefinder.group import Group, GroupKey
from filefinder.matches import Matches, get_groups_indices

logger = logging.getLogger(__name__)


class _FilterUserFunc(Protocol):
    """Defines the signature of filters callables (for static type checkers).

    See `<https://mypy.readthedocs.io/en/stable/protocols.html#callback-protocols>`__
    for details.
    """

    def __call__(
        self, __finder: "Finder", filename: str, matches: Matches, **kwargs: Any
    ) -> bool:
        ...


FilterPartial = Callable[["Finder", str, Matches], bool]


class Finder:
    """Find files using a filename pattern.

    The Finder object is the main entrance point to this library.
    Given a root directory and a filename pattern, it can search for all
    corresponding files.

    Parameters
    ----------
    root:
        The root directory of the filetree where all files can be found.
    pattern:
        A regular expression with the addition of 'groups'. See :ref:`find-files` for
        details.
    use_regex:
        If True, characters outside of groups are considered as valid regex (and
        not escaped). Default is False.
    scan_everything
        If true, look into all sub-directories up to a depth of :attr:`max_scan_depth` .
        This is appropriate if the pattern contains optional sub-directories. If false
        (default), check that every sub-directory matches its part of the regular
        expression, thus avoiding some work.
    """

    max_scan_depth: int = 32
    """Maximum sub-directory depth to scan when :attr:`scan_everything` is True."""

    def __init__(
        self,
        root: str,
        pattern: str,
        use_regex: bool = False,
        scan_everything: bool = False,
    ):
        self.root: str = root
        """The root directory of the finder."""
        self.use_regex: bool = use_regex
        """If True, characters outside of groups are considered as valid regex
        (and not escaped). Default is False."""

        self.scan_everything: bool = scan_everything
        """Whether to scan all subdirectories."""

        self._pattern: str

        self.groups: list[Group] = []
        self._segments: list[str] = []
        """Segments of the pattern. Used to replace specific groups.
        `['text before group 1', 'group 1',
        'text before group 2, 'group 2', ...]`
        """
        self._files: list[tuple[str, Matches]] = []
        self.scanned: bool = False
        """True if files have been scanned with current parameters.

        Is reset to False if the cache (of scanned files) is voided, for instance by
        operation like changing fixed values of groups.
        """

        self.filters: list[FilterPartial] = []
        """List of filters to apply to found files.

        Each value is a :func:`partial function<functools.partial>` with keyword
        arguments applied.
        """

        self.set_pattern(pattern)

    @property
    def n_groups(self) -> int:
        """Number of groups in pre-regex."""
        return len(self.groups)

    @property
    def files(self) -> list[tuple[str, Matches]]:
        """List of filenames and their matches.

        Will scan files when accessed and cache the result, if it has not
        already been done.
        """
        if not self.scanned:
            self.find_files()
        return self._files

    def __repr__(self) -> str:
        """Human readable information."""
        return "\n".join([super().__repr__(), self.__str__()])

    def __str__(self) -> str:
        """Human readable information."""
        s = [
            f"root: {self.root}",
            f"pattern: {self._pattern}",
            f"regex: {self.get_regex()}",
        ]

        fixed_groups = [
            (i, g.fixed_value)
            for i, g in enumerate(self.groups)
            if g.fixed_value is not None
        ]
        if fixed_groups:
            s.append("fixed groups:")
            s += [f"\t fixed #{i} to {v}" for i, v in fixed_groups]

        if not self.scanned:
            s.append("not scanned")
        else:
            s.append(f"scanned: found {len(self._files)} files")
        return "\n".join(s)

    def set_scan_everything(self, scan_everything: bool, /) -> None:
        """Set value for attribute :attr:`scan_everything`.

        Void cache if necessary.
        """
        if scan_everything != self.scan_everything:
            self.scan_everything = scan_everything
            self._void_cache()

    def set_use_regex(self, use_regex: bool, /) -> None:
        """Set value for attribute :attr:`use_regex`."""
        self.use_regex = use_regex

    def get_group_names(self, fixed: bool | None = None) -> set[str]:
        """Get the names of groups in the pattern.

        Parameters
        ----------
        fixed
            If True, only return names of groups with a fixed value. If False, return
            only those without a fixed value. If None (default), return for all groups.
        """
        groups = self.groups
        if fixed is not None:
            groups = [g for g in groups if g.fixed == fixed]

        return set(g.name for g in groups)

    def get_files(
        self,
        relative: bool = False,
        nested: Sequence[str | Sequence[str]] | None = None,
    ) -> list:
        """Return files that matches the regex.

        Lazily scan files: if files were already scanned, just return
        the stored list of files.
        Scanned files are flushed if the regex is changed (by fixing group
        for instance).

        Parameters
        ----------
        relative:
            If True, filenames are returned relative to the finder
            root directory. If not, paths are absolute (default).
        nested:
            If not None, return nested list of filenames with each level
            corresponding to a group, or set of group. Last set in the list
            is at the innermost level.

        Raises
        ------
        KeyError: A group name in `nested` is not found in the pattern.
        """

        def get_files(files_matches):
            if relative:
                return [f for f, _ in files_matches]
            return [self.get_absolute(f) for f, _ in files_matches]

        def get_key(matches: Matches, level: list[str]) -> str:
            return ":".join(
                [
                    match.get_match(parse=False)
                    for match in matches
                    if match.group.name in level
                ]
            )

        def nest(files_matches, levels, relative):
            if len(levels) == 0:
                return get_files(files_matches)

            level = levels[0]
            files_grouped = []
            matches: dict[str, int] = {}
            # We need to sort files by their value.
            # We use all unparsed matches joined in a single string as a key
            # (using get_key). We store it in a dictionnary, the value being
            # the corresponding index
            for f, m in files_matches:
                key = get_key(m, level)
                if key not in matches:
                    matches[key] = len(matches)
                    files_grouped.append([])
                files_grouped[matches[key]].append((f, m))

            return [nest(grp, levels[1:], relative) for grp in files_grouped]

        if not self.scanned:
            self.find_files()

        if nested is None:
            files = get_files(self._files)
        else:
            names = set(g.name for g in self.groups)
            nested = [[name] if isinstance(name, str) else name for name in nested]
            for name in itertools.chain(*nested):
                if name not in names:
                    raise KeyError(f"{name} is not in Finder groups.")
            files = nest(self._files, nested, relative)

        return files

    def get_relative(self, filename: str) -> str:
        """Get filename path relative to root."""
        return os.path.relpath(filename, self.root)

    def get_absolute(self, filename: str) -> str:
        """Get absolute path to filename."""
        return os.path.join(self.root, filename)

    def fix_group(self, key: GroupKey, value: str | Any, fix_discard: bool = False):
        """Fix a group to a string.

        Parameters
        ----------
        key:
            Can be the index of a group in the pattern (starts at 0), or the
            name of a group. If multiple groups share the same name, they are
            all fixed to the same value.
        value:
            Will replace the match for all files. Can be a string, or a value
            that will be formatted using the group format string.
            A list of values will be joined by the regex '|' OR.
            A string will be interpreted as a regular expression, so all special
            characters should be properly escaped.
        fix_discard:
            If True, groups with the 'discard' option will still be fixed.
            Default is False.
        """
        for m in self.get_groups(key):
            if not fix_discard and m.discard:
                continue
            m.fix_value(value)
        self._void_cache()

    def fix_groups(
        self,
        fixes: dict[Any, str | Any] | None = None,
        fix_discard: bool = False,
        **fixes_kw: str | Any,
    ):
        """Fix multiple groups at once.

        Parameters
        ----------
        fixes:
            Dictionnary of `{group key: value}`. See :func:`fix_group` for
            details.
        fix_discard:
            If True, groups with the 'discard' option will still be fixed.
            Default is False.
        fixes_kw:
            Same as `fixes`. Takes precedence.
        """
        if fixes is None:
            fixes = {}
        fixes.update(**fixes_kw)
        for f in fixes.items():
            self.fix_group(*f, fix_discard=fix_discard)

    def unfix_groups(self, *keys: str):
        """Unfix groups.

        Parameters
        ----------
        keys:
           Keys to find groups to unfix. See :func:`get_groups`.
           If no key is provided, all groups will be unfixed.
        """
        if not keys:
            for g in self.groups:
                g.unfix()
        else:
            for key in keys:
                groups = self.get_groups(key)
                for g in groups:
                    g.unfix()
        self._void_cache()

    def find_matches(self, filename: str, relative: bool = True) -> Matches | None:
        """Find matches for a given filename.

        Apply regex to `filename` and return the results as a :class:`~.matches.Matches`
        object. Fixed values are applied as normal.

        Parameters
        ----------
        filename:
            Filename to retrieve matches from.
        relative:
            True if the filename is relative to the finder root directory
            (default). If False, the filename is made relative before being
            matched.

        Returns
        -------
        matches
            A :class:`~.matches.Matches` object, or None if the filename did not match.
        """
        if not relative:
            filename = self.get_relative(filename)

        return Matches.from_filename(filename, self.get_regex(), self.groups)

    def make_filename(
        self,
        fixes: dict | None = None,
        relative: bool = False,
        **kw_fixes: Any,
    ) -> str:
        """Return a filename.

        Replace groups with provided values.
        All groups must be fixed prior, or with `fixes` argument.

        Parameters
        ----------
        fixes:
            Dictionnary of fixes (group name or index: value). For details, see
            :func:`fix_group`. Will (temporarily) supplant group fixed
            prior. If prior fix is a list, first item will be used.
        relative:
            If the filename should be relative to the finder root directory.
            Default is False.
        kw_fixes:
            Same as `fixes`. Takes precedence.

        Raises
        ------
        ValueError: `use_regex` is activated.
        """
        if self.use_regex:
            raise ValueError(
                "Cannot generate a valid filename if regex "
                "is present outside groups (`use_regex=True`)."
            )

        if fixes is None:
            fixes = {}
        fixes.update(**kw_fixes)

        segments = self._segments.copy()
        groups = [copy(g) for g in self.groups]  # shallow copy (no reparsing of def)

        for i, g in enumerate(groups):
            if g.name in fixes:
                g.fix_value(fixes[g.name])
            if i in fixes:
                g.fix_value(fixes[i])

            if g.fixed_string is not None:
                segments[2 * i + 1] = g.fixed_string
            else:
                raise ValueError(f"Group '{g!s}' has no fixed value.")

        filename = "".join(segments).replace("/", os.sep)

        if not relative:
            filename = self.get_absolute(filename)

        return filename

    def get_pattern(self) -> str:
        """Get filename pattern."""
        return self._pattern

    def set_pattern(self, pattern: str):
        """Set pattern and parse for group objects."""
        self._void_cache()
        self._pattern = pattern
        groups_starts = [m.start() + 1 for m in re.finditer(r"%\(", pattern)]

        # This finds the matching end parenthesis for each group start
        self.groups = []
        splits = [0]  # separation between groups
        for idx, start in enumerate(groups_starts):
            end = None
            level = 1
            for i, c in enumerate(pattern[start + 1 :]):
                if c == "(":
                    level += 1
                elif c == ")":
                    level -= 1
                    if level == 0:  # matching parenthesis
                        end = start + i + 1
                        break

            if end is None:  # did not find matching parenthesis :(
                end = start + 6
                substr = pattern[start - 1 : end]
                if end < len(self._pattern):
                    substr += "..."
                raise ValueError(f"No group end found for '{substr}'")

            self.groups.append(Group(pattern[start + 1 : end], idx))
            splits += [start - 1, end + 1]  # -1 removes the %

        self._segments = [pattern[i:j] for i, j in zip(splits, splits[1:] + [None])]

    def _get_regex(self) -> str:
        segments = self._segments.copy()
        if not self.use_regex:
            # escape regex outside groups
            segments = [
                s if (i % 2 == 1) else re.escape(s) for i, s in enumerate(segments)
            ]

        for idx, group in enumerate(self.groups):
            segments[2 * idx + 1] = group.get_regex()

        return "".join(segments)

    def get_regex(self) -> str:
        """Return regex."""
        return self._get_regex().replace("/", re.escape(os.sep))

    def get_regex_subdirs(self) -> list[str]:
        """Return regexes for each sub-directory."""
        return self._get_regex().split("/")

    def add_filter(self, func: _FilterUserFunc, **kwargs: Any):
        """Add a filter with which to select scanned files.

        See :doc:`/filtering` for details.

        :rtype func: Callable

        Parameters
        ----------
        func: ~collections.abc.Callable[[Finder, str, Matches, ...], bool]
            Callable that returns True if the file is to be kept, False otherwise.
        kwargs
            Will be passed to the function when executed.
        """
        filt = functools.partial(func, **kwargs)
        self.filters.append(filt)
        self._apply_filters(filt)

    def clear_filters(self) -> None:
        """Remove all filters."""
        self.filters.clear()
        self._void_cache()

    def _apply_filters(self, filter: FilterPartial | None = None) -> None:
        filters = [filter] if filter is not None else self.filters
        kept_files = []
        for filename, matches in self._files:
            if all(filt(self, filename, matches) for filt in filters):
                kept_files.append((filename, matches))
        self._files = kept_files

    def find_files(self) -> None:
        """Find files to scan and store them.

        Is automatically called when accessing :attr:`files` or :func:`get_files`. Apply
        all filters and sort files alphabetically.
        """
        if self.scan_everything:
            self._find_files_scan_everything()
        else:
            self._find_files_subdirectories()

        if self.filters:
            self._apply_filters()

        self._files.sort(key=lambda x: x[0])

    def _find_files_scan_everything(self) -> None:
        """Find files checking every sub-directory.

        Because having to check if a sub-directory matches the pattern is difficult,
        this allows for more exotic patterns where a folder separator can appear in a
        capturing group, by example for optional sub-directories.

        This will the whole filetree under :attr:`root_directory` and check every file
        found, which can be significant work in some cases.
        """
        pattern = re.compile(self.get_regex())
        files_matched = []

        for dirpath, dirnames, filenames in os.walk(self.root):
            depth = dirpath.rstrip(os.sep).count(os.sep) - self.root.rstrip(
                os.sep
            ).count(os.sep)
            if depth > self.max_scan_depth:
                dirnames.clear()

            for f in filenames:
                to_root = self.get_relative(os.path.join(dirpath, f))
                matches = Matches.from_filename(to_root, pattern, self.groups)
                if matches is not None:
                    files_matched.append((to_root, matches))

        self.scanned = True
        self._files = files_matched

    def _find_files_subdirectories(self) -> None:
        """Find files checking sub-directories along the way.

        Each sub-directory must match against its corresponding part of the generated
        regular expression. This is ill suited if any group contains a folder
        separator. But it will limit the number of sub-directories to explore and
        thus the number of files to check.
        """
        max_log_lines = 3

        subpatterns = [re.compile(rgx) for rgx in self.get_regex_subdirs()]
        files = []
        for dirpath, dirnames, filenames in os.walk(self.root):
            depth = dirpath.rstrip(os.sep).count(os.sep) - self.root.rstrip(
                os.sep
            ).count(os.sep)
            pattern = subpatterns[depth]

            if depth == len(subpatterns) - 1:
                dirnames.clear()  # look no deeper
                files += [
                    self.get_relative(os.path.join(dirpath, f)) for f in filenames
                ]
            elif logger.isEnabledFor(logging.DEBUG):
                dirlogs = dirnames[:max_log_lines]
                if len(dirnames) > max_log_lines:
                    dirlogs += ["..."]
                logger.debug(
                    "depth: %d, pattern: %s, folders:\n\t%s",
                    depth,
                    pattern.pattern,
                    "\n\t".join(dirlogs),
                )

            # Removes directories not matching regex
            to_remove = [d for d in dirnames if not pattern.fullmatch(d)]
            for d in to_remove:
                dirnames.remove(d)

        logger.debug("Found %s files in sub-directories", len(files))

        # Now only retain files that match the full pattern
        pattern = re.compile(self.get_regex())
        files_matched = []

        for f in files:
            matches = Matches.from_filename(f, pattern, self.groups)
            if matches is not None:
                files_matched.append((f, matches))

        if logger.isEnabledFor(logging.DEBUG):
            filelogs = files[:max_log_lines]
            if len(files) > max_log_lines:
                filelogs += ["..."]
            logger.debug(
                "regex: %s, files:\n\t%s", pattern.pattern, "\n\t".join(filelogs)
            )
            logger.debug("Found %s matching files in %s", len(files_matched), self.root)

        self.scanned = True
        self._files = files_matched

    def _void_cache(self) -> None:
        self.scanned = False
        self._files.clear()

    def get_groups(self, key: GroupKey) -> list[Group]:
        """Return list of groups corresponding to key.

        Parameters
        ----------
        key: int, str, or list of int
            Can be group index, name, or group+name combination with the
            syntax: 'group:name'.

        Returns
        -------
        List of groups corresponding to key.

        Raises
        ------
        KeyError: No group found.
        TypeError: Key type is not valid.
        """
        selected = get_groups_indices(self.groups, key)
        groups = [self.groups[i] for i in selected]
        return groups



"""Matches management."""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck

import logging
import re
import typing as t
from collections import abc

from .group import Group, GroupKey

logger = logging.getLogger(__name__)


class Match:
    """Match extract from a filename.

    Parameters
    ----------
    group
        Group used to get this match.
    match
        Match object for the complete filename.
    idx
        Index of the group in the match object.

    Attributes
    ----------
    group:
        Group used to get this match.
    match_str:
        String matched in the filename.
    start:
        Start index of match in the filename.
    end:
        End index of match in the filename.
    match_parsed:
        Parsed value. None if parsing was not successful.
    """

    def __init__(self, group: Group, match: re.Match, idx: int):
        self.group: Group = group
        self.match_str: str = match.group(idx + 1)
        self.start: int = match.start(idx + 1)
        self.end: int = match.end(idx + 1)

    def __repr__(self):
        """Human readable information."""
        return "\n".join([super().__repr__(), self.__str__()])

    def __str__(self):
        """Human readable information."""
        return f"{self.group!s} = {self.match_str}"

    @property
    def match_parsed(self) -> t.Any | None:
        """Parsed value, None if parsing was not successful."""
        try:
            return self.group.parse(self.match_str)
        except Exception:
            logger.debug("Failed to parse for group %s", str(self.group))
            return None

    def get_match(self, parse: bool = True) -> t.Any:
        """Get match string or value.

        Parameters
        ----------
        parse:
            If True (default) return the parsed value instead of the matched string.

        Raises
        ------
        ValueError: Could not parse the match.
        """
        if parse:
            if self.match_parsed is None:
                raise ValueError(
                    f"Failed to parse value '{self.match_str}' "
                    f"for group '{self.group!s}'."
                )
            return self.match_parsed
        return self.match_str


class Matches:
    """Scan an input file and store the results.

    Parameters
    ----------
    match
        Match object obtained from a filename. It should have as much capturing groups
        as the pattern.
    groups
        Sequence of Groups objects present in the pattern.

    """

    @classmethod
    def from_filename(
        cls, filename: str, pattern: re.Pattern | str, groups: abc.Sequence[Group]
    ) -> "Matches | None":
        """Find matches for a given filename.

        Parameters
        ----------
        filename:
            Filename to retrieve matches from.
        pattern
            Compiled match pattern to use. If left to None, we generate the current
            regex.

        Returns
        -------
        matches
            A :class:`Matches` object, or None if the filename did not match.

        Raises
        ------
        IndexError
            Not as many matches as groups. Maybe one of the group regex contains an
            additional (unwanted) capturing group ?
        """
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        m = pattern.fullmatch(filename)
        if m is None:
            return None

        if len(groups) != len(m.groups()):
            raise IndexError(
                "Not as many captured matches as pattern groups. "
                "Does one of the group regex contains a capturing group ?"
            )

        return cls(m, groups)

    def __init__(self, match: re.Match, groups: abc.Sequence[Group]):
        self.matches: list[Match] = []
        """Matches for a single filename."""
        self.groups = list(groups)
        """Groups used."""

        assert len(match.groups()) == len(groups)

        for i in range(len(groups)):
            self.matches.append(Match(groups[i], match, i))

    def __repr__(self) -> str:
        """Human readable information."""
        return "\n".join([super().__repr__(), self.__str__()])

    def __str__(self) -> str:
        """Human readable information."""
        return "\n".join([str(m) for m in self.matches])

    def __getitem__(self, key: GroupKey) -> t.Any:
        """Get first parsed value corresponding to key.

        Ignore groups with the 'discard' option.
        """
        return self.get_value(key, parse=True, keep_discard=False)

    def __iter__(self) -> abc.Iterator[Match]:
        """Iterate over matches."""
        return iter(self.matches)

    def __len__(self) -> int:
        """Return number of matches."""
        return len(self.matches)

    def get_values(
        self, key: GroupKey, parse: bool = True, keep_discard: bool = False
    ) -> list[t.Any]:
        """Get matched values corresponding to key.

        Return a list of values, even if only one group is selected.

        Parameters
        ----------
        key:
            Group(s) to select, either by index or name.
        parse:
            If True (default), return the parsed value. If False return the
            matched string.
        keep_discard:
            If true groups with the 'discard' option are kept. Defauult is false.
        """
        matches = self.get_matches(key, keep_discard=keep_discard)
        values = [m.get_match(parse=parse) for m in matches]
        return values

    def get_value(
        self, key: GroupKey, parse: bool = True, keep_discard: bool = False
    ) -> t.Any:
        """Get matched value corresponding to key.

        Return a single value. If multiple groups correspond to ``key``,
        the value of the first one to appear in the pattern is returned.

        Parameters
        ----------
        key:
            Group(s) to select, either by index or name.
        parse:
            If True (default), return the parsed value. If False return the
            matched string.
        keep_discard:
            If true groups with the 'discard' option are kept. Defauult is false.

        Raises
        ------
        KeyError: No group with no 'discard' option was found.
        """
        values = self.get_values(key, parse=parse, keep_discard=keep_discard)
        if len(values) == 0:
            raise KeyError(
                "No group was found " f"(key: {key}, keep_discard: {keep_discard})"
            )
        if len(values) > 1:
            if any(v != values[0] for v in values[1:]):
                logger.warning(
                    "Different parsed values for key %s (%s)", str(key), repr(values)
                )
        return values[0]

    def get_matches(self, key: GroupKey, keep_discard: bool = False) -> list[Match]:
        """Get Match objects corresponding to key.

        Parameters
        ----------
        key:
            Group(s) to select, either by index or name.
        keep_discard:
            If true groups with the 'discard' option are kept. Defauult is false.

        Returns
        -------
        List of Match corresponding to the key.
        """
        selected = get_groups_indices(self.groups, key)
        matches = [self.matches[k] for k in selected]
        if not keep_discard:
            matches = [m for m in matches if not m.group.discard]
        return matches


def get_groups_indices(groups: list[Group], key: GroupKey) -> list[int]:
    """Get sorted list of groups indices corresponding to key.

    Key can be an integer index, or a string of a group name. Since multiple
    groups can share the same name, multiple indices can be returned (sorted).

    Raises
    ------
    IndexError: No group found corresponding to the key
    TypeError: Key is not int or str
    """
    if isinstance(key, int):
        return [key]
    if isinstance(key, str):
        selected = [i for i, group in enumerate(groups) if group.name == key]

        if len(selected) == 0:
            raise IndexError(f"No group found for key '{key}'")
        return selected

    raise TypeError("Key must be int or str.")



"""Group management."""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck

from __future__ import annotations

import logging
import re
from typing import Any

from .format import FormatAbstract, get_format

logger = logging.getLogger(__name__)


GroupKey = int | str
"""Can be used to select one or more groups in a pattern."""

"""On Group parsing:

explain reasoning behind PATTERN, and alternatives explored
"""


class GroupParseError(Exception):
    """Custom errors when parsing group definition."""

    def __init__(self, message: str, group: Group | None = None):
        if group is not None:
            message += f" ({group.definition})"
        super().__init__(message)


class Group:
    """Manage a group inside the filename pattern.

    Parameters
    ----------
    definition:
        Group definition.
    idx:
        Index of the group in the filename pattern.

    Raises
    ------
    GroupParseError: Invalid group definition.
    """

    PATTERN = re.compile(
        "(?P<name>[^:]+?)(?:(?P<fmt>:fmt=.+?)|(?P<rgx>:rgx=.*?)"
        "|(?P<bool>:bool=.*?(?::.*?)??)"
        "|(?P<opt>:opt)|(?P<discard>:discard)){,5}"
    )
    """Pattern used to find properties in group definition."""

    DEFAULT_GROUPS = {
        "I": [r"\d+", "d"],  # index
        "Y": [r"\d{4}", "04d"],  # year
        "m": [r"\d\d", "02d"],  # month
        "d": [r"\d\d", "02d"],  # day
        "j": [r"\d{3}", "03d"],  # dayofyear
        "H": [r"\d\d", "02d"],  # hour
        "M": [r"\d\d", "02d"],  # minute
        "S": [r"\d\d", "02d"],  # second
        "x": [r"%Y%m%d", "08d"],  # date
        "X": [r"%H%M%S", "06d"],  # time
        "F": [r"%Y-%m-%d", "s"],  # formated date
        "B": [r"[a-zA-Z]*", "s"],  # month / month abbreviation
        "text": [r"\w", "s"],
        "char": [r"\S*", "s"],
    }
    """Regex and format strings for various default groups.

    See the :ref:`name` section of documentation for details.
    """

    def __init__(self, definition: str, idx: int):
        self.definition = definition
        """The string that created the group ``%(definition)``."""
        self.idx: int = idx
        """Index inside the pre-regex."""

        self.name: str = ""
        """Group name."""
        self.rgx: str = ""
        """Regex."""
        self.fmt: FormatAbstract = get_format("s")
        """Format string object."""
        self.discard: bool = False
        """If the group should not be used when retrieving values from matches."""
        self.options: tuple[str, str] | None = None
        """Tuple of the two possibilities indicated by the full ``:bool``
        specification, in order (False, True), so that a simple getitem works."""
        self.optional: bool = False
        """If True, the whole group is marked as optional (``()?``).
        Is set to False unless specification ':opt' is indicated."""

        self._fixed = False
        self.fixed_value: Any | None = None
        self.fixed_string: str | None = None
        self.fixed_regex: str | None = None

        self._parse_group_definition()

    def _parse_group_definition(self) -> None:
        """Parse group definition against a regex to retrieve specs."""
        m = self.PATTERN.fullmatch(self.definition)
        if m is None:
            raise GroupParseError("Could not parse the group definition.", self)
        self._check_duplicates(m)
        specs = m.groupdict()

        self.name = specs["name"]

        # Set to defaults if name is known
        default = self.DEFAULT_GROUPS.get(self.name)
        if default is not None:
            self.rgx, fmt_def = default
            self.fmt = get_format(fmt_def)

        # Extract specs
        for k in ["rgx", "fmt", "bool"]:
            if specs[k] is not None:
                specs[k] = specs[k].removeprefix(f":{k}=")
        rgx = specs["rgx"]
        fmt = specs["fmt"]
        bol = specs["bool"]

        # Flags
        self.discard = specs["discard"] is not None
        self.optional = specs["opt"] is not None

        # Override default format
        if fmt:
            self.fmt = get_format(fmt)
            if not rgx:  # No need to generate rgx if it is provided
                self.rgx = self.fmt.generate_expression()

        # Boolean format
        self.options = None
        if bol is not None:
            options = bol.split(":", maxsplit=1)
            if len(options) == 1:
                options.append("")
            self.options = tuple(options[::-1])
            self.rgx = "|".join(re.escape(s) for s in options)

        # Override regex
        if rgx:
            self.rgx = rgx

        if not self.rgx:
            raise GroupParseError("No regex has been produced.", self)

        if self.options is not None:
            # rgx is A|B, A and B being strings we don't do "% replacement"
            return
        self.rgx = self._replace_regex_defaults(self.rgx)

    def _check_duplicates(self, m: re.Match):
        """Check if the definition does not contain duplicates.

        The matching pattern (:attr:`PATTERN`) is written so that specs (rgx, fmt, ...)
        can be given in any order, while still matching the full string.

        The pattern is made of every possible spec in a OR list, which can be repeated
        up to 5 times. Regex only keep the last captured group. We must be a bit sly
        to check duplicates. For that I check that the groups we have account for
        the whole string. If part of the string is not in our match, something has
        been overwritten.
        """
        accounted_for: list[tuple[int, int]] = []

        for k, v in m.groupdict().items():
            if v is not None:
                accounted_for.append((m.start(k), m.end(k)))

        accounted_for.sort(key=lambda x: x[0])

        # to make sure we get to the end
        n = len(self.definition)
        accounted_for.append((n, n))

        pos = 0
        for span in accounted_for:
            if pos == span[0]:
                pos = span[1]
            else:
                raise GroupParseError(
                    (
                        "The specs found do not account for the full definition. "
                        "There is most likely a duplicate spec."
                    ),
                    self,
                )

    def _replace_regex_defaults(self, regex: str) -> str:
        """Recursively replace defaults regexes of the form ``%[a-zA-Z]``.

        Replacements are taken from :attr:`Group.DEFAULT_GROUPS`.

        A '%' in the regex should be escaped by another: '%%'.

        Raises
        ------
        KeyError: Unknown replacement.
        """

        def replace(match: re.Match):
            group = match.group(1)
            if group == "%":
                return "%"
            if group in self.DEFAULT_GROUPS:
                replacement = self.DEFAULT_GROUPS[group][0]
                if "%" in replacement:  # need to go recursive
                    return self._replace_regex_defaults(replacement)
                return replacement
            raise KeyError(f"Unknown replacement '{match.group(0)}'.")

        return re.sub("%([a-zA-Z%])", replace, regex)

    def __repr__(self) -> str:
        """Human readable information."""
        return "\n".join([super().__repr__(), self.__str__()])

    def __str__(self) -> str:
        """Human readable information."""
        return f"{self.name}:{self.idx:d}"

    @property
    def fixed(self) -> bool:
        """True if the group has fixed value(s)."""
        return self._fixed

    def format(self, value: Any) -> str:
        """Return formatted string from value."""
        return self.fmt.format(value)

    def parse(self, string: str) -> Any:
        """Return parsed value from string."""
        # parsing boolean
        if self.options is not None:
            if string == self.options[0]:
                return False
            elif string == self.options[1]:
                return True
            else:
                raise ValueError(
                    f"Cannot parse '{string}' into boolean from options {self.options}"
                )

        return self.fmt.parse(string)

    def fix_value(self, fix: Any | bool | str):
        """Fix the group regex to a specific value.

        Parameters
        ----------
        fix:
            A string is directly used as a regular expression, otherwise the
            value is formatted according to the group 'format' specification.
        """
        self._fixed = True
        self.fixed_value = fix

        if not isinstance(fix, list | tuple):
            fix = [fix]

        if len(fix) == 0:
            raise ValueError("A list of fixes must contain at least one element.")

        strings = []
        regexes = []
        for f in fix:
            # if a string, leave it as is
            if isinstance(f, str):
                out = f
                rgx = f
            # if optional A|B choice
            elif isinstance(f, bool):
                if self.options is None:
                    raise ValueError(
                        f"{self.name} group has no A|B options, "
                        "cannot fix value with a boolean."
                    )
                out = self.options[f]
                rgx = re.escape(out)
            else:
                out = self.format(f)
                rgx = re.escape(out)
            strings.append(out)
            regexes.append(rgx)

        self.fixed_string = strings[0]
        self.fixed_regex = "|".join(regexes)

    def unfix(self):
        """Unfix value."""
        self._fixed = False
        self.fixed_value = None
        self.fixed_string = None
        self.fixed_regex = None

    def get_regex(self) -> str:
        """Get group regex.

        Returns the fixed value if previously specified.
        Insert the regex into a capturing group, and make it optional if
        the ``:opt`` was indicated
        """
        if self.fixed_regex is not None:
            rgx = self.fixed_regex
        else:
            rgx = self.rgx

        # Make it matching
        rgx = f"({rgx})"

        if self.optional is True:
            rgx += "?"

        return rgx



"""Generate regex from string format, and parse strings.

Parameters of the format-string are retrieved.
See `Format Mini Language Specification
<https://docs.python.org/3/library/string.html#formatspec>`__.

Thoses parameters are then used to generate a regular expression, or to parse
a string formed from the format.

Only 's', 'd', 'f', 'e' and 'E' formats types are supported.

The width of the format string is not respected when matching with a regular
expression.
"""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck

import re
from collections.abc import Mapping
from typing import Any

FORMAT_REGEX = (
    r"((?P<fill>.)?(?P<align>[<>=^]))?"
    r"(?P<sign>[-+ ])?"
    r"(?P<z>z)?(?P<alternate>#)?(?P<zero>0)?"
    r"(?P<width>\d+?)?"
    r"(?P<grouping>[,_])?"
    r"(?P<precision>\.\d+?)?"
    r"(?P<type>[a-zA-Z])"
)
"""The regular expression used to parse a format string.

Follows the Format Specification Mini-Language.
``[[fill]align][sign]["z"]["#"]["0"][width][grouping_option]["." precision][type]``
"""
FORMAT_PATTERN = re.compile(FORMAT_REGEX)


class FormatError(Exception):
    """Error related to Format object."""


class FormatParsingError(FormatError):
    """Could not parse a format-string."""


class DangerousFormatError(FormatError):
    """Dangerous format-string leading to ambiguities."""


class InvalidFormatTypeError(FormatError):
    """Unsupported type of format-string."""


class FormatValueParsingError(FormatError):
    """Could not parse value."""


class FormatAbstract:
    """Represent a format string.

    Can generate an appropriate regular expression corresponding to that format string
    (to some limitations), generate a string from a value, or parse such a string into
    a value.

    Users are not meant to instanciate those objects directly, use :func:`get_format`
    instead (or its alias for retro-compatibility :func:`Format`).

    Parameters
    ----------
    fmt:
        Format string.
    params:
        Mapping of options/parameters of the format mini-language to their values.
        (type, fill, align, sign, alternate, zero, width, grouping, precision).
        They should not contain None values.
    """

    ALLOWED_TYPES = "sdfeE"

    def __init__(self, fmt: str, params: Mapping[str, Any]):
        self.fmt: str = fmt

        self.type: str = params["type"]
        self.fill: str = params["fill"]
        self.align: str = params["align"]
        self.sign: str = params["sign"]
        self.alternate: bool = params["alternate"]
        self.zero: bool = params["zero"]
        self.width: int = params["width"]
        self.grouping: str = params["grouping"]
        self.precision: int = params["precision"]

        if self.type not in self.ALLOWED_TYPES:
            raise InvalidFormatTypeError(
                f"Invalid format type '{type}', expected one of {self.ALLOWED_TYPES}."
            )

    def format(self, value: Any) -> str:
        """Return formatted string of a value."""
        return f"{{:{self.fmt}}}".format(value)

    def get_fill_regex(self):
        """Return regex for matching fill characters."""
        return f"{re.escape(self.fill)}*"

    def add_outer_alignement(self, rgx: str) -> str:
        """Add necessary regex for alignement characters.

        If width is not specified, does nothing.
        """
        if self.width == 0 or self.align == "=":
            return rgx

        out = [rgx]
        fill_rgx = self.get_fill_regex()

        # value is right-aligned or center
        if self.align in ">^":
            out.insert(0, fill_rgx)

        # value is left-aligned or center
        if self.align in "<^":
            out.append(fill_rgx)

        return "".join(out)

    def parse(self, s: str) -> Any:
        """Parse string generated with this format into an appropriate value."""
        raise NotImplementedError()

    def generate_expression(self, capture=False) -> str:
        """Generate a regular expression matching strings created with this format.

        Parameters
        ----------
        capture
            If true, add capturing groups that will be used to parse the value by
            selecting only relevant information. Default is false.
        """
        raise NotImplementedError()


class FormatString(FormatAbstract):
    """Represent a format string for strings (type s)."""

    ALLOWED_TYPES = "s"

    type = "s"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.align == "=":
            raise FormatError("'=' alignement not allowed for string format.")
        if self.alternate:
            raise FormatError("Alternate form (#) not allowed for string format.")
        if self.grouping:
            raise FormatError("Grouping not allowed for string format.")
        if self.sign:
            raise FormatError("Sign not allowed for string format.")

    def parse(self, s: str) -> str:
        """Parse string generated with this format into an appropriate value."""
        pattern = self.generate_expression(capture=True)
        m = re.fullmatch(pattern, s)
        if m is None:
            raise FormatValueParsingError(
                f"Error parsing '{s}' with pattern '{pattern}'"
            )
        return m.group(1)

    def generate_expression(self, capture=False) -> str:
        """Generate a regular expression matching strings created with this format."""
        rgx = ".*?"
        if capture:
            rgx = f"({rgx})"
        rgx = self.add_outer_alignement(rgx)
        return rgx


class FormatNumberAbstract(FormatAbstract):
    """Represent a format string for numbers (type d, f, e, E)."""

    ALLOWED_TYPES = "dfeE"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Reject dubious formats
        digits = list(map(str, range(10)))
        if self.width > 0 and (
            (self.fill in digits and self.align in "<^")
            or (self.fill in digits[1:] and self.align in "=")
            or (self.fill in digits and self.align == ">" and self.sign == "-")
            or (self.fill == "-" and self.align in ">^=" and self.sign == "-")
        ):
            raise DangerousFormatError(
                f"Dangerous combination of fill character ({self.fill}), "
                f"alignement ({self.align}) and sign ({self.sign}) for "
                f"format ({self.fmt})"
            )

    def prepare_parse(self, s: str) -> str:
        """Remove special characters.

        Remove characters that throw off int() and float() parsing: fill/alignment
        characters and grouping symbols.

        Returns
        -------
        s
            a string ready to be casted to the appropriate type.
        """
        pattern = self.generate_expression(capture=True)
        m = re.fullmatch(pattern, s)
        if m is None:
            raise FormatValueParsingError(
                f"Error parsing '{s}' with pattern '{pattern}'"
            )
        # join all capturing groups (sign, number)
        s = "".join(m.groups(""))
        # remove grouping characters
        s = re.sub("[,_]", "", s)
        return s

    def get_left_of_decimal(self) -> str:
        """Get regex for the numbers left of decimal point.

        Will deal with grouping if present.
        """
        if self.grouping:
            rgx = rf"\d?\d?\d(?:{self.grouping}\d{{3}})*"
        else:
            rgx = r"\d+"
        return rgx

    def get_sign_regex(self, capture=False) -> str:
        """Get sign regex with approprite zero padding."""
        if self.sign == "-":
            rgx = "-?"
        elif self.sign == "+":
            rgx = r"[+-]"
        elif self.sign == " ":
            rgx = r"[\s-]"
        else:
            raise KeyError("Sign not in {+- }") from FormatError

        if capture:
            rgx = f"({rgx})"

        # padding is added between sign and numbers
        if self.width > 0 and self.align == "=":
            rgx = rgx + self.get_fill_regex()
        return rgx


class FormatInteger(FormatNumberAbstract):
    """Represent a format string for integers (type d)."""

    ALLOWED_TYPES = "d"

    def parse(self, s: str) -> int:
        """Parse string generated with format."""
        s = self.prepare_parse(s)
        return int(s)

    def generate_expression(self, capture=False) -> str:
        """Generate regex from format string."""
        rgx = self.get_sign_regex(capture=capture)
        number = self.get_left_of_decimal()
        if capture:
            number = f"({number})"
        rgx += number
        rgx = self.add_outer_alignement(rgx)
        return rgx


class FormatFloat(FormatNumberAbstract):
    """Represent a format string for floats (type f, e, E)."""

    ALLOWED_TYPES = "feE"

    def get_right_of_decimal(self) -> str:
        """Return regex for numbers after decimal points.

        Including the decimal point itself. It will respect 'alternate' option and
        the specified precision.
        """
        rgx = ""
        if self.precision != 0 or self.alternate:
            rgx += r"\."
        if self.precision != 0:
            rgx += rf"\d{{{self.precision:d}}}"
        return rgx

    def parse(self, s: str) -> float:
        """Parse string generated with format."""
        s = self.prepare_parse(s)
        return float(s)

    def get_left_of_decimal(self) -> str:
        """Get regex for the numbers left of decimal point.

        Will deal with grouping if present. Some simplifications for eE formats.
        Only use groupings for '0=' alignment and enforce single digits grouping.
        """
        if self.type == "f":
            return super().get_left_of_decimal()
        if self.grouping and self.fill == "0" and self.align == "=":
            return rf"0?0?\d(?:{self.grouping}00\d)*"
        return r"\d"

    def generate_expression(self, capture=False) -> str:
        """Generate a regular expression matching strings created with this format."""
        if self.type == "f":
            rgx = self.get_sign_regex(capture=capture)
            number = self.get_left_of_decimal() + self.get_right_of_decimal()
            if capture:
                number = f"({number})"
            rgx += number

            rgx = self.add_outer_alignement(rgx)
            return rgx

        assert self.type in "eE"
        rgx = self.get_sign_regex(capture=capture)
        number = (
            self.get_left_of_decimal()
            + self.get_right_of_decimal()
            + rf"{self.type}[+-]\d{{2,3}}"
        )
        if capture:
            number = f"({number})"
        rgx += number
        rgx = self.add_outer_alignement(rgx)
        return rgx


FORMAT_CLASSES: dict[str, type[FormatAbstract]] = dict(
    s=FormatString,
    d=FormatInteger,
    f=FormatFloat,
    e=FormatFloat,
    E=FormatFloat,
)


def get_format(format: str) -> FormatAbstract:
    """Parse format parameters and return appropriate Format object."""
    m = FORMAT_PATTERN.fullmatch(format)
    if m is None:
        raise FormatParsingError(f"Format-string '{format}' not valid.")
    params = m.groupdict()

    kind = params["type"]

    # fill boolean parameters
    params["alternate"] = params["alternate"] == "#"
    params["zero"] = params["zero"] == "0"

    # special case
    if params["zero"] and kind in "dfeE":
        if params["fill"] is None:
            params["fill"] = "0"
        if params["align"] is None:
            params["align"] = "="

    # TODO Precision not supported in s kind (it truncates the value)
    if kind == "s" and params["precision"]:
        raise FormatError("Precision parameter is currently not supported.")

    # defaults values for unset remaining parameters
    defaults = dict(align="<", fill=" ")
    if kind in "dfeE":
        defaults |= dict(align=">", sign="-", width="0", precision=".6", grouping="")

    for k, v in defaults.items():
        if params[k] is None:
            params[k] = v

    if kind in "dfeE":
        # convert to correct kind
        params["width"] = int(params["width"])
        params["precision"] = int(params["precision"].removeprefix("."))

    if kind not in FORMAT_CLASSES:
        raise InvalidFormatTypeError(
            f"Invalid format kind '{kind}', "
            f"expected one of '{list(FORMAT_CLASSES.keys())}'."
        )
    return FORMAT_CLASSES[kind](format, params)


# Retrocompatible alias
Format = get_format



"""Functions to retrieve values from filename."""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck

import datetime as dt
import logging
from collections.abc import Callable

from .finder import Finder
from .matches import Matches

logger = logging.getLogger(__name__)


def get_date(matches: Matches, default_date: dict | None = None) -> dt.datetime:
    """Retrieve date from matched elements.

    If a matcher is *not* found in the filename, it will be replaced by the
    element of the default date argument.
    Matchers that can be used are (in order of increasing priority):
    YBmdjHMSFxX. If two matchers have the same name, the last one in the
    pre-regex will get priority.

    Parameters
    ----------
    matches:
        Matches obtained from a filename.
    default_date:
        Default date. Dictionnary with keys: year, month, day, hour, minute,
        and second. Defaults to 1970-01-01 00:00:00
    """
    name_to_datetime = dict(
        Y="year", m="month", d="day", H="hour", M="minute", S="second"
    )

    def get_elts(elts: dict[str, str], names: str, callback: Callable):
        for name in names:
            elt = elts.pop(name, None)
            if elt is not None:
                date.update(callback(elt, name))

    def process_int(elt: str, name: str) -> dict[str, int]:
        return {name_to_datetime[name]: int(elt)}

    def process_month_name(elt: str, name: str) -> dict[str, int]:
        return dict(month=_find_month_number(elt))

    def process_doy(elt: str, name: str) -> dict[str, int]:
        d = dt.datetime(date["year"], 1, 1) + dt.timedelta(days=int(elt) - 1)
        return dict(month=d.month, day=d.day)

    date = {"year": 1970, "month": 1, "day": 1, "hour": 0, "minute": 0, "second": 0}

    if default_date is None:
        default_date = {}
    date.update(default_date)

    elts = {
        m.group.name: m.get_match(parse=False) for m in matches if not m.group.discard
    }

    elts_needed = set("xXYmdBjHMSF")
    if len(set(elts.keys()) & elts_needed) == 0:
        logger.warning(
            "No matchers to retrieve a date from." " Returning default date."
        )

    # Process month name first to keep element priorities simples
    get_elts(elts, "B", process_month_name)

    # Decompose elements
    elt = elts.pop("F", None)
    if elt is not None:
        elts["Y"] = elt[:4]
        elts["m"] = elt[5:7]
        elts["d"] = elt[8:10]

    elt = elts.pop("x", None)
    if elt is not None:
        elts["Y"] = elt[:4]
        elts["m"] = elt[4:6]
        elts["d"] = elt[6:8]

    elt = elts.pop("X", None)
    if elt is not None:
        elts["H"] = elt[:2]
        elts["M"] = elt[2:4]
        if len(elt) > 4:  # noqa: PLR2004
            elts["S"] = elt[4:6]

    # Process elements
    get_elts(elts, "Ymd", process_int)
    get_elts(elts, "j", process_doy)
    get_elts(elts, "HMS", process_int)

    return dt.datetime(**date)  # type: ignore


def _find_month_number(name: str) -> int:
    """Find a month number from its name.

    Name can be the full name (January) or its three letter abbreviation (jan).
    The casing does not matter.
    """
    names = [
        "january",
        "february",
        "march",
        "april",
        "may",
        "june",
        "july",
        "august",
        "september",
        "october",
        "november",
        "december",
    ]
    names_abbr = [c[:3] for c in names]

    name = name.lower()
    if name in names:
        return names.index(name) + 1
    if name in names_abbr:
        return names_abbr.index(name) + 1

    raise ValueError(f"Could not interpret month name '{name}'")


def filter_by_range(
    finder: Finder,
    filename: str,
    matches: Matches,
    group: str,
    min: float | None = None,
    max: float | None = None,
) -> bool:
    """Filter filename using the value parsed for `group`.

    Keep filename for which the value parsed for `group` fall within a specific range
    defined by `min` and `max`.

    Parameters
    ----------
    group
        Name of the group to use the parsed value. The first non-discard group of that
        name will be used.
    min
        If not None, the parsed value must be above this.
    max
        If not None, the parsed value mest be below this.

    Raises
    ------
    TypeError
        `min` and `max` cannot be both None.
    """
    if min is None and max is None:
        raise TypeError("`min` and `max` cannot be both None.")

    parsed = matches.get_value(group, parse=True, keep_discard=False)

    if min is not None and parsed < min:
        return False
    if max is not None and parsed > max:
        return False
    return True


def filter_date_range(
    finder: Finder,
    filename: str,
    matches: Matches,
    start: dt.date | str,
    stop: dt.date | str,
    default_date: dict | None = None,
) -> bool:
    """Filter filename to be between two dates.

    Parameters
    ----------
    start, stop
        Start and stop dates that define the range of dates to keep. Can each be a
        :class:`datetime.date` or :class:`datetime.datetime` object; or a string in
        which case a datetime object is created with
        :meth:`~datetime.datetime.fromisoformat`.
    default_date
        Is passed to :func:`get_date`.

    Returns
    -------
    keep
        True if the file is within the range and must be kept. False otherwise.
    """
    if isinstance(start, str):
        start = dt.datetime.fromisoformat(start)
    if isinstance(stop, str):
        stop = dt.datetime.fromisoformat(stop)

    if start >= stop:
        raise ValueError(f"Start ({start}) must be before stop ({stop})")

    current = get_date(matches, default_date=default_date)

    return start <= current <= stop
